from ..object_classes.text_object import TextObject
import numpy as np


class Text(TextObject):
    def __init__(self, text, position=np.array([0, 0, 0]), color=(0, 0, 0), size=1):
        super().__init__(text, position, color, size)



from ..object_classes.flat_faces_object import FlatFacesObject
import numpy as np


class Fcube(FlatFacesObject):
    def __init__(self, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0), size=1):
        vertices = np.array(
            [
                np.array([0, 0, 0]),
                np.array([size, 0, 0]),
                np.array([size, size, 0]),
                np.array([0, size, 0]),
                np.array([0, 0, size]),
                np.array([size, 0, size]),
                np.array([size, size, size]),
                np.array([0, size, size]),
            ],
            dtype=float,
        )

        faces = [
            ([0, 1, 2, 3], (255, 111, 100)),
            ([4, 5, 6, 7], (255, 111, 65)),
            ([0, 1, 5, 4], (255, 65, 100)),
            ([1, 2, 6, 5], (65, 111, 100)),
            ([2, 3, 7, 6], (255, 189, 100)),
            ([3, 0, 4, 7], (255, 111, 200)),
        ]

        super().__init__(vertices, faces, position, color)

        self.move_relative(np.array([-size / 2, -size / 2, -size / 2]))



import numpy as np


def average_points(points):
    """
    Averages a list of points
    :param points: list of points
    :return: average point
    """
    return np.mean(points, axis=0, dtype=float)



from .src.sapphirerenderer.main import SapphireRenderer, get_pygame_object
from .src.sapphirerenderer.utility_objects.particle_manager import ParticleManager

from .src.sapphirerenderer.object_classes.base_object import Object
from .src.sapphirerenderer.object_classes.flat_faces_object import FlatFacesObject
from .src.sapphirerenderer.object_classes.text_object import TextObject
from .src.sapphirerenderer.object_classes.wireframe_object import WireframeObject



from ..object_classes.base_object import Object
import numpy as np
import pygame
from ..settings import (
    draw_vertices,
    draw_lines,
    line_thickness,
    point_thickness,
)
from ..point_math.project_point import project_point
from ..point_math.matricies import get_pitch_yaw_roll_matrix
from ..point_math.average_points import average_points


class WireframeObject(Object):
    def __init__(
        self,
        vertices,
        lines,
        position=np.array([0, 0, 0], dtype=float),
        color=(0, 0, 0),
    ):
        super().__init__(position=position, color=color)
        self.original_vertices = vertices.copy()
        self.vertices = vertices
        self.lines = lines
        self.position = np.array([0, 0, 0], dtype=float)
        self.color = color
        self.rotation = np.array([0, 0, 0], dtype=float)

        self.center_point = average_points(self.vertices)

        self.drawing = False
        self.ambiguous = False

        self.move_absolute(position)

        self.show()

    def move_relative(self, vector):
        """
        Move the object by a relative amount
        :param vector: the amount to move by
        :return:
        """
        self._wait_for_draw()

        self.ambiguous = True
        self.position += vector
        for i in range(len(self.vertices)):
            self.vertices[i] += vector
        self.center_point = average_points(self.vertices)
        self.ambiguous = False

    def move_absolute(self, vector):
        """
        Move the object to an absolute position
        :param vector: the position to move to
        :return:
        """
        self._wait_for_draw()

        self.ambiguous = True
        vector = np.array(vector, dtype=float)
        self.position = vector
        for i in range(len(self.vertices)):
            self.vertices[i] = self.original_vertices[i] + vector
        self.center_point = average_points(self.vertices)
        self.ambiguous = False

    def __rotate(self, x_axis, y_axis, z_axis):
        self._wait_for_draw()

        self.ambiguous = True
        rotation_matrix = get_pitch_yaw_roll_matrix(x_axis, z_axis, y_axis)
        self.vertices = np.dot(self.vertices, rotation_matrix.T)
        self.original_vertices = np.dot(self.original_vertices, rotation_matrix.T)
        self.rotation += np.array([x_axis, z_axis, y_axis], dtype=float)
        self.rotation_matrix = get_pitch_yaw_roll_matrix(*self.rotation)
        self.ambiguous = False

    def rotate_local(self, x_axis, y_axis, z_axis):
        """
        Rotate the object around a point
        :param x_axis: x axis rotation in degrees
        :param y_axis: y axis rotation in degrees
        :param z_axis: z axis rotation in degrees
        :param point: the point to rotate around
        :return:
        """

        self.rotate_around_point(x_axis, y_axis, z_axis, self.center_point)

    def rotate_around_point(
        self, x_axis, y_axis, z_axis, point=np.array([0, 0, 0], dtype=float)
    ):
        """
        Rotate the object around a point
        :param x_axis: x axis rotation in degrees
        :param y_axis: y axis rotation in degrees
        :param z_axis: z axis rotation in degrees
        :param point: the point to rotate around
        :return:
        """

        # convert to radians
        x_axis, y_axis, z_axis = (
            np.radians(x_axis),
            np.radians(y_axis),
            np.radians(z_axis),
        )

        self._wait_for_draw()

        self.ambiguous = True
        self.vertices -= point
        self.__rotate(x_axis, y_axis, z_axis)
        self.vertices += point
        self.center_point = average_points(self.vertices)

        self.ambiguous = False

    def __str__(self):
        return self.__class__.__name__

    def draw(self, surface, camera):
        self._wait_for_ambiguous()

        moved_vertices = self.vertices - camera.position
        reshaped_vertices = moved_vertices.reshape(-1, 1, moved_vertices.shape[1])
        rotated_vertices = np.sum(camera.rotation_matrix * reshaped_vertices, axis=-1)

        projected_vertices = [
            project_point(
                vertex,
                camera.offset_array,
                camera.focal_length,
            )
            for vertex in rotated_vertices
        ]

        if self.lines is not None and draw_lines:
            for line in self.lines:
                start, s_scale = projected_vertices[line[0]]
                end, e_scale = projected_vertices[line[1]]

                if start is None or end is None:
                    continue

                pygame.draw.line(
                    surface,
                    line[2] if len(line) > 2 else self.color,
                    start,
                    end,
                    max(int(line_thickness * (s_scale + e_scale) / 2), 1),
                )
        if draw_vertices:
            for vertex in projected_vertices:
                vertex, scale = vertex

                if draw_vertices and vertex is not None:
                    pygame.draw.circle(
                        surface,
                        self.color,
                        vertex,
                        max(int(point_thickness * scale), 1),
                    )



import os
from .utility_objects.camera import Camera
import numpy as np
from .settings import (
    camera_move_speed,
    camera_rotate_speed,
    fps,
    show_fps,
    lock_fps,
)
from time import time
import threading
from .point_math.project_point import project_point
from .object_classes.flat_faces_object import get_face_distances
import pygame

average_fps_list = []


def get_pygame_object():
    return pygame


class SapphireRenderer:
    def __init__(
        self, width=1000, height=1000, draw_axis=False, movement_handling=True
    ):
        """
        Initialize the renderer
        :param width: Width of the window
        :param height: Height of the window
        :param draw_axis: Draws the axis lines, use-full for debugging
        ;param: movement_handling: Whether to handle movement of the camera
        """
        self.display = None

        self.movement_handling = movement_handling

        self.width = width
        self.height = height

        self.camera_rotate_speed = camera_rotate_speed
        self.camera_move_speed = camera_move_speed

        self.camera = Camera(self, position=np.array((0.0, -3.0, 0.0)))

        self.loaded_objects = []
        self.instance_objects = []
        self.load_objects()

        if draw_axis:
            self.add_object("Axes")

        self.running = True

        self.thread = threading.Thread(target=self.render_loop)
        self.thread.start()

    def load_objects(self):
        # go through all files in objects and load them
        for file in os.listdir(os.path.dirname(__file__) + "/objects"):
            if file.endswith(".py") and file != "__init__.py":
                try:
                    exec(f"from .objects.{file[:-3]} import *")
                    obj_class_name = f"{file[:1].upper().replace('_', '')}{file[1:-3].replace('_', '')}"
                    self.loaded_objects.append((obj_class_name, eval(obj_class_name)))
                except Exception as e:
                    print(f"Failed to load object {file}: {e}")

    def add_object(self, obj_name, args=None):
        """
        Adds an object to the scene
        :param obj_name: The class name of the object
        :param args: The args to pass to the init of the class
        :return: returns the object created
        """
        for obj_class_name, obj_class in self.loaded_objects:
            if obj_class_name == obj_name:
                obj = obj_class(*args) if args is not None else obj_class()
                self.instance_objects.append(obj)
                return obj

    def direct_add_object(self, obj):
        """
        Adds an object to the scene
        :param obj: The object to add
        :return:
        """
        self.instance_objects.append(obj)
        return obj

    def remove_object(self, obj):
        """
        Removes an object from the scene
        :param obj: The object to remove
        :return:
        """
        self.instance_objects.remove(obj)

    def update(self):
        self.camera.update()
        for obj in self.instance_objects:
            obj.update()

    def user_input(self, pygame, scale_factor=1.0):
        # wasd to move camera
        keys = pygame.key.get_pressed()
        # if shift is pressed, move faster
        if keys[pygame.K_LSHIFT]:
            scale_factor *= 2

        if keys[pygame.K_w]:
            self.camera.move_relative((self.camera_move_speed * scale_factor, 0, 0))
        if keys[pygame.K_s]:
            self.camera.move_relative((-self.camera_move_speed * scale_factor, 0, 0))
        if keys[pygame.K_a]:
            self.camera.move_relative((0, self.camera_move_speed * scale_factor, 0))
        if keys[pygame.K_d]:
            self.camera.move_relative((0, -self.camera_move_speed * scale_factor, 0))
        if keys[pygame.K_q]:
            self.camera.move_relative((0, 0, -self.camera_move_speed * scale_factor))
        if keys[pygame.K_e]:
            self.camera.move_relative((0, 0, self.camera_move_speed * scale_factor))

        if keys[pygame.K_LEFT]:
            self.camera.rotate_relative((0, -self.camera_rotate_speed * scale_factor))
        if keys[pygame.K_RIGHT]:
            self.camera.rotate_relative((0, self.camera_rotate_speed * scale_factor))
        if keys[pygame.K_UP]:
            self.camera.rotate_relative((-self.camera_rotate_speed * scale_factor, 0))
        if keys[pygame.K_DOWN]:
            self.camera.rotate_relative((self.camera_rotate_speed * scale_factor, 0))

    def compiled_draw(self, surface, camera):
        """
        Draw the compiled objects were all verts and faces are put together, sorted, and drawn
        :param surface: the pygame surface to draw on
        :param camera: the camera to draw from
        :return:
        """
        compiled_faces = []
        compiled_verts = []
        index_offset = 0

        # if number of objects is 0 with compiled verts enabled, return
        if len([obj for obj in self.instance_objects if obj.compile_verts]) == 0:
            return

        for obj in self.instance_objects:
            if not obj.is_hidden() and obj.compile_verts:
                obj.drawing = True
                obj.wait_for_ambiguous()

                compiled_verts.extend(obj.vertices)

                # append if the object has shadow, the strength of shadow, and the reverse rotation matrix to each face
                for face in obj.faces:
                    face = (
                        [vertex + index_offset for vertex in face[0]],
                        face[1],
                        face[2],
                    )

                    compiled_faces.append(
                        face
                        + (
                            obj.shadow,
                            obj.shadow_effect,
                            obj.negative_rotation_matrix,
                        )
                    )
                obj.drawing = False
                index_offset += len(obj.vertices)

        if len(compiled_faces) == 0:
            return

        face_distances = get_face_distances(
            compiled_faces, compiled_verts, camera.position
        )

        sorted_indices = np.argsort(face_distances)[
            ::-1
        ]  # Sorting indices in descending order

        compiled_faces = [compiled_faces[i] for i in sorted_indices]

        moved_vertices = compiled_verts - camera.position
        reshaped_vertices = moved_vertices.reshape(-1, 1, moved_vertices.shape[1])
        rotated_vertices = np.sum(camera.rotation_matrix * reshaped_vertices, axis=-1)

        projected_vertices = [
            project_point(
                vertex,
                camera.offset_array,
                camera.focal_length,
            )[0]
            for vertex in rotated_vertices
        ]

        for face in compiled_faces:
            face_verts = face[0]
            face_color = face[1] if len(face) > 1 else (0, 0, 0)
            face_normal = face[2] if len(face) > 2 else None
            face_shadow = face[3] if len(face) > 3 else False
            shadow_effect = face[4] if len(face) > 4 else 1
            negative_rotation_matrix = face[5] if len(face) > 5 else None

            # rotate face normal by object rotation
            if face_normal is not None and face_shadow:
                face_normal = np.dot(face_normal, negative_rotation_matrix)

                shadow_normal = ((face_normal[2] + 255) / 510) * 255

                shadow_normal /= shadow_effect

                # if shadow_normal is nan, set it to 255
                if np.isnan(shadow_normal):
                    shadow_normal = 255

                # dim the color based on the shadow_normal
                face_color = tuple(
                    int(color * shadow_normal / 255) for color in face_color
                )

            if any(
                vertex is None
                for vertex in [projected_vertices[vertex] for vertex in face_verts]
            ):
                continue
            pygame.draw.polygon(
                surface,
                face_color,
                [projected_vertices[vertex] for vertex in face_verts],
            )

    def render_loop(self):
        import pygame

        self.display = pygame.display.set_mode((self.width, self.height))
        self.display.fill((255, 255, 255))
        pygame.display.set_caption("Sapphire Renderer")

        while self.running:
            frame_start = time() + 0.00001

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False

            self.display.fill((255, 255, 255))
            self.update()

            # sort objects by distance from camera, reverse so that objects closer to camera are drawn last
            self.instance_objects.sort(
                key=lambda obj: np.linalg.norm(obj.position - self.camera.position),
                reverse=True,
            )

            for obj in self.instance_objects:
                if not obj.is_hidden() and not obj.compile_verts:
                    obj.draw(self.display, self.camera)

            self.compiled_draw(self.display, self.camera)

            pygame.display.flip()

            # if fps is higher than fps setting, wait
            if lock_fps and time() - frame_start < 1 / fps:
                pygame.time.wait(int(1000 * (1 / fps - (time() - frame_start))))

            real_fps = 1 / (time() - frame_start)
            average_fps_list.append(real_fps)

            average_fps = sum(average_fps_list) / len(average_fps_list)

            if len(average_fps_list) > 10:
                average_fps_list.pop(0)

            if show_fps:
                pygame.display.set_caption(
                    f"Sapphire Renderer - FPS: {int(average_fps)}"
                )

            if self.movement_handling:
                self.user_input(pygame, fps / real_fps)

        pygame.quit()

    def stop(self):
        self.running = False
        self.thread.join()



import numpy as np
from ..object_classes.flat_faces_object import FlatFacesObject


def load_obj(filename):
    vertices = []
    faces = []

    with open(filename, "r") as file:
        for line in file:
            if line.startswith("v "):
                vertex = [float(coord) for coord in line.split()[1:]]
                vertices.append(vertex)
            elif line.startswith("f "):
                face_data = [
                    [int(index.split("/")[0]) - 1 for index in line.split()[1:]]
                ]
                faces.append(face_data)

    return np.array(vertices), faces


class Fobj(FlatFacesObject):
    def __init__(
        self,
        filename,
        position=np.array([0.0, 0.0, 0.0]),
        color=(0, 0, 0),
        random_color=False,
    ):
        if not filename.endswith(".obj"):
            raise ValueError("File must be an OBJ file")

        # Load OBJ file
        vertices, faces = load_obj(filename)

        for face in faces:
            if not random_color:
                face.append(color)
            else:
                face.append(np.random.randint(0, 255, 3))

        super().__init__(vertices, faces, position, color, True)



from sapphirerenderer import SapphireRenderer
import numpy as np
from time import sleep


def main():
    renderer = SapphireRenderer(draw_axis=True)

    large_cube = renderer.add_object(
        "Cube", args=(np.array([0, 0, 0], dtype=float), (0, 255, 0), 1)
    )
    small_cube = renderer.add_object(
        "Cube", args=(np.array([4, 0, 0], dtype=float), (255, 0, 0), 0.5)
    )

    large_text = renderer.add_object(
        "Text", args=("Large Cube", np.array([0, 0, 0]), (0, 0, 0))
    )

    small_text = renderer.add_object(
        "Text", args=("Small Cube", np.array([4, 0, 0]), (0, 0, 0), 0.5)
    )

    large_vector = np.array([0, 0, 0], dtype=float)
    small_vector = np.array([0, 0.05, 0.05], dtype=float)

    large_mass = 1
    small_mass = 0.5
    gravitational_constant = 9.81

    while renderer.running:
        gravitational_force = (
            gravitational_constant
            * large_mass
            * small_mass
            / np.linalg.norm(large_cube.center_point - small_cube.center_point) ** 2
        )

        large_acceleration = gravitational_force / large_mass
        small_acceleration = gravitational_force / small_mass

        # make vector that points from large cube to small cube with a magnitude of large acceleration
        large_accel_vector = (
            (small_cube.center_point - large_cube.center_point)
            * large_acceleration
            / 1000
        )
        small_accel_vector = (
            (large_cube.center_point - small_cube.center_point)
            * small_acceleration
            / 1000
        )

        large_vector += large_accel_vector
        small_vector += small_accel_vector

        # large_cube.move_relative(large_vector)
        small_cube.move_relative(small_vector)

        small_text.set_position(small_cube.center_point)

        sleep(0.01)


if __name__ == "__main__":
    main()



import numpy as np
from stl import mesh
from ..object_classes.flat_faces_object import FlatFacesObject


class Fstl(FlatFacesObject):
    def __init__(
        self,
        filename,
        position=np.array([0.0, 0.0, 0.0]),
        color=(0, 0, 0),
        random_color=False,
        shadow=True,
    ):
        # Load STL file
        mesh_data = mesh.Mesh.from_file(filename)

        # Extract vertices
        vertices = np.unique(mesh_data.vectors.reshape(-1, 3), axis=0)

        # Extract faces
        faces = []
        for triangle in mesh_data.vectors:
            face_vertices = []
            for vertex in triangle:
                index = np.nonzero((vertices == vertex).all(axis=1))[0][0]
                face_vertices.append(index)
            (
                faces.append([face_vertices, color])
                if not random_color
                else faces.append([face_vertices, np.random.randint(0, 255, 3)])
            )

        normals = mesh_data.normals
        for i, face in enumerate(faces):
            face.append(normals[i] / np.linalg.norm(normals[i]) * 255)

        faces = [tuple(face) for face in faces]

        super().__init__(vertices, faces, position, color, shadow)



import numpy as np
from time import sleep
from ..point_math.matricies import get_pitch_yaw_roll_matrix


class Object:
    def __init__(
        self,
        position=np.array([0, 0, 0]),
        rotation=np.array((0, 0, 0)),
        scale=1,
        color=(0, 0, 0),
    ):
        """
        Base object class
        :param position: the position of the object
        :param rotation: the rotation of the object
        :param scale: the scale of the object
        :param color: the color of the object
        """
        self.position = position
        self.rotation = rotation
        self.scale = scale
        self.hidden = True
        self.color = color

        self.drawing = False
        self.ambiguous = False

        self.rotation_matrix = get_pitch_yaw_roll_matrix(*rotation)

    def wait_for_draw(self):
        while self.drawing:
            sleep(0.0001)

    def wait_for_ambiguous(self):
        while self.ambiguous:
            sleep(0.0001)

    def is_hidden(self):
        return self.hidden

    def hide(self):
        self.hidden = True
        return self

    def show(self):
        self.hidden = False
        return self

    def set_position(self, position):
        self.position = position

    def move_relative(self, vector):
        self.position += vector

    def set_rotation(self, rotation):
        self.rotation = rotation

    def set_scale(self, scale):
        self.scale = scale

    def set_color(self, color):
        self.color = color

    def get_position(self):
        return self.position

    def get_rotation(self):
        return self.rotation

    def get_scale(self):
        return self.scale

    def get_color(self):
        return self.color

    def update(self):
        """
        empty update for child classes to override
        :return:
        """
        pass

    def __str__(self):
        return self.__class__.__name__



from ..object_classes.base_object import Object
import pygame
from ..point_math.average_points import average_points
import numpy as np
from ..point_math.project_point import project_point
from ..point_math.matricies import get_pitch_yaw_roll_matrix
from numba import njit

pygame.init()


@njit(fastmath=True)
def get_vertex_distances(vertices, camera_position):
    return [np.linalg.norm(vertex - camera_position) for vertex in vertices]


def get_face_distances(faces, vertices, camera_position):
    vertex_distances = get_vertex_distances(vertices, camera_position)
    face_distances = []
    for face in faces:
        face_distance = 0
        for vertex_index in face[0]:
            face_distance += vertex_distances[vertex_index]
        face_distances.append(face_distance / len(face[0]))
    return face_distances


class FlatFacesObject(Object):
    def __init__(
        self,
        vertices,
        faces,
        position=np.array([0, 0, 0]),
        color=(0, 0, 0),
        shadow=False,
        shadow_effect=1,
        compile_verts=True,
    ):
        """
        Object with flat faces
        :param vertices: the vertices of the object
        :param faces: the faces of the object
        :param position: the position of the object
        :param color: the color of the object
        :param shadow: whether to render shadows
        :param shadow_effect: the strength of the shadow
        :param compile_verts: whether to compile the vertices in the rendering loop
        """
        self.drawing = False
        self.ambiguous = False
        self.compile_verts = compile_verts

        self.position = position
        super().__init__(color=color, position=self.position)
        self.show()

        self.vertices = vertices
        self.__util_move_to_zero()
        self.original_vertices = vertices.copy()

        self.faces = faces
        self.shadow_effect = shadow_effect
        self.shadow = shadow

        self.rotation = np.array([0, 0, 0], dtype=float)
        self.negative_rotation_matrix = get_pitch_yaw_roll_matrix(*-self.rotation)
        self.center_point = average_points(vertices)

        self.move_absolute(position)

    def move_relative(self, vector):
        """
        Move the object by a relative amount
        :param vector: the amount to move by
        :return:
        """
        self.wait_for_draw()

        self.ambiguous = True
        self.position += vector
        for i in range(len(self.vertices)):
            self.vertices[i] += vector
        self.center_point = average_points(self.vertices)
        self.ambiguous = False

    def __util_move_to_zero(self):
        """
        Move the object to an absolute position
        :param vector: the position to move to
        :return:
        """
        self.wait_for_draw()

        self.ambiguous = True
        self.center_point = average_points(self.vertices)
        self.position = np.array([0, 0, 0], dtype=float)
        for i in range(len(self.vertices)):
            self.vertices[i] -= self.center_point
        self.ambiguous = False

    def move_absolute(self, vector):
        """
        Move the object to an absolute position
        :param vector: the position to move to
        :return:
        """
        self.wait_for_draw()

        self.ambiguous = True
        vector = np.array(vector, dtype=float)
        self.position = vector
        for i in range(len(self.vertices)):
            self.vertices[i] = self.original_vertices[i] + vector
        self.center_point = average_points(self.vertices)
        self.ambiguous = False

    def __rotate(self, x_axis, y_axis, z_axis):
        self.wait_for_draw()

        self.ambiguous = True
        rotation_matrix = get_pitch_yaw_roll_matrix(x_axis, z_axis, y_axis)
        self.vertices = np.dot(self.vertices, rotation_matrix.T)
        self.original_vertices = np.dot(self.original_vertices, rotation_matrix.T)
        self.rotation += np.array([x_axis, z_axis, y_axis], dtype=float)
        self.negative_rotation_matrix = get_pitch_yaw_roll_matrix(*-self.rotation)
        self.ambiguous = False

    def rotate_local(self, x_axis, y_axis, z_axis):
        """
        Rotate the object around its center point
        :param x_axis: x axis rotation in degrees
        :param y_axis: y axis rotation in degrees
        :param z_axis: z axis rotation in degrees
        :return:
        """
        # convert to radians
        x_axis, y_axis, z_axis = (
            np.radians(x_axis),
            np.radians(y_axis),
            np.radians(z_axis),
        )

        self.wait_for_draw()
        self.ambiguous = True
        self.vertices -= self.center_point
        self.__rotate(x_axis, y_axis, z_axis)
        self.vertices += self.center_point
        self.ambiguous = False

    def rotate_around_point(
        self, x_axis, y_axis, z_axis, point=np.array([0, 0, 0], dtype=float)
    ):
        """
        Rotate the object around a point
        :param x_axis: x axis rotation in degrees
        :param y_axis: y axis rotation in degrees
        :param z_axis: z axis rotation in degrees
        :param point: the point to rotate around
        :return:
        """

        # convert to radians
        x_axis, y_axis, z_axis = (
            np.radians(x_axis),
            np.radians(y_axis),
            np.radians(z_axis),
        )

        self.wait_for_draw()

        self.ambiguous = True
        self.vertices -= point
        self.__rotate(x_axis, y_axis, z_axis)
        self.vertices += point
        self.ambiguous = False

    def set_scale(self, scale_factor, center_point=None):
        """
        Scale the object
        :param scale_factor: the factor to scale by
        :param center_point: the point to scale around
        :return:
        """
        if center_point is None:
            center_point = self.center_point

        self.wait_for_draw()

        self.ambiguous = True
        self.vertices -= center_point
        self.vertices *= scale_factor
        self.vertices += center_point

        self.original_vertices -= center_point
        self.original_vertices *= scale_factor
        self.original_vertices += center_point
        self.ambiguous = False

    def draw(self, surface, camera):
        """
        Draw the object
        :param surface: the pygame surface to draw on
        :param camera: the camera to draw from
        :return:
        """
        self.wait_for_ambiguous()
        self.drawing = True

        face_distances = get_face_distances(self.faces, self.vertices, camera.position)

        sorted_indices = np.argsort(face_distances)[
            ::-1
        ]  # Sorting indices in descending order

        self.faces = [self.faces[i] for i in sorted_indices]

        moved_vertices = self.vertices - camera.position
        reshaped_vertices = moved_vertices.reshape(-1, 1, moved_vertices.shape[1])
        rotated_vertices = np.sum(camera.rotation_matrix * reshaped_vertices, axis=-1)

        projected_vertices = [
            project_point(
                vertex,
                camera.offset_array,
                camera.focal_length,
            )[0]
            for vertex in rotated_vertices
        ]

        for face in self.faces:
            face_verts = face[0]
            face_color = face[1] if len(face) > 1 else self.color
            face_normal = face[2] if len(face) > 2 else None

            # rotate face normal by object rotation
            if face_normal is not None and self.shadow:
                face_normal = np.dot(face_normal, self.negative_rotation_matrix)

                shadow_normal = ((face_normal[2] + 255) / 510) * 255

                shadow_normal /= self.shadow_effect

                # if shadow_normal is nan, set it to 255
                if np.isnan(shadow_normal):
                    shadow_normal = 255

                # dim the color based on the shadow_normal
                face_color = tuple(
                    int(color * shadow_normal / 255) for color in face_color
                )

            if any(
                vertex is None
                for vertex in [projected_vertices[vertex] for vertex in face_verts]
            ):
                continue
            pygame.draw.polygon(
                surface,
                face_color,
                [projected_vertices[vertex] for vertex in face_verts],
            )

        self.drawing = False



from ..object_classes.wireframe_object import WireframeObject
import numpy as np


class Line(WireframeObject):
    def __init__(self, position=np.array([0.0, 0.0, 0.0]), color=(0, 0, 0)):
        vertices = np.array(
            [
                [0, 0, 0],
                [0, 0, 0],
            ],
            dtype=float,
        )

        lines = [[0, 1, color]]

        super().__init__(vertices, lines, position, color)

    def change_color(self, color):
        self.lines[0][2] = color
        self.color = color

    def change_vertices(self, vertices):
        self.vertices = vertices

    def change_start(self, start):
        self.vertices[0] = start

    def change_end(self, end):
        self.vertices[1] = end



from ..object_classes.base_object import Object
import numpy as np
import pygame
from ..point_math.project_point import project_point

pygame.init()


class TextObject(Object):
    def __init__(self, text, position, color=(0, 0, 0), size=1):
        """
        Object to display text
        :param text: the text to display
        :param position: the position of the text
        :param color: the color of the text
        :param size: the size of the text
        """
        super().__init__(position=position, color=color)
        self.text = text
        self.size = size
        self.show()

    def set_text(self, text):
        """
        Set the text
        :param text: the new text
        :return:
        """
        self.text = text

    def get_text(self):
        """
        Get the text
        :return: the text
        """
        return self.text

    def draw(self, surface, camera):
        """
        Draw the text
        :param surface: the pygame surface to draw on
        :param camera: the camera to draw from
        :return:
        """
        # draw text with left corner at position, should also be scaled based on distance from camera
        camera_distance = np.linalg.norm(self.position - camera.position)

        moved_vertices = self.position.copy() - camera.position
        rotated_vertices = np.sum(camera.rotation_matrix * moved_vertices, axis=-1)

        flat_position = project_point(
            rotated_vertices,
            camera.offset_array,
            camera.focal_length,
        )[0]

        if flat_position is not None:

            # scale text based on distance from camera
            scale = 1 / camera_distance

            font = pygame.font.Font("freesansbold.ttf", int(100 * scale * self.size))
            text = font.render(self.text, True, self.color)

            # clamp flat position to be between -10000 and 10000
            flat_position = np.clip(flat_position, -10000, 10000)

            surface.blit(text, flat_position)
